#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <stdbool.h>

#define STB_TRUETYPE_IMPLEMENTATION
#include <stb/std_truetype.h>

#define ED_GFX_IMPLEMENTATION
#define ED_STRING_IMPLEMENTATION
#define ED_HT_IMPLEMENTATION
#define ED_UI_IMPLEMENTATION
#define ED_BUFFER_IMPLEMENTATION
#include "string.h"
#include "ht.h"
#include "ui.h"
#include "ed_array.h"
#include "gfx.h"
#include "buffer.h"

// static Arena default_arena = {0};
// static Arena temporary_arena = {0};
// static Arena *context_arena = &default_arena;

void *context_alloc(size_t size) {
    // assert(context_arena);
    // return arena_alloc(context_arena, size);
    return malloc(size);
}

typedef struct {
    float position[2];
    float size[2];
    float border_size[2];
} GpuUiRect;
arrayTemplate(GpuUiRect);

typedef struct {
    float atlas_position[2];
    float size[2];
    float position[2];
    float y_offset;
    float _haha_alignment;
} GpuGlyph;
arrayTemplate(GpuGlyph);

typedef struct {
    float screen_size[2];
} GpuUniformParams;

static struct {
    array(GpuUiRect) gpu_ui_rects;
    array(GpuGlyph) gpu_glyphs;
    array(GpuGlyph) glyph_cache;

    bool should_exit;

    ui_context ui_cx;
    gfx_context_t *gfx_cx;

    ed_buffer test_buffer;
} state;

void queue_text(string text, float position[2]) {
    float x = 0;
    for (size_t i=0; i < text.len; ++i) {
        if (text.data[i] >= 32) {
            GpuGlyph glyph = state.glyph_cache.data[text.data[i] - 32];

            glyph.position[0] = x+position[0];
            glyph.position[1] = position[1];
            x += glyph.size[0]/2+4;

            pushArray(GpuGlyph, &state.gpu_glyphs, glyph);
        }
    }
}

void queue_ui_rect(float position[2], float size[2], float border_size) {
    GpuUiRect rect = (GpuUiRect) {
        .position = { position[0], position[1] },
        .size = { size[0], size[1] },
        .border_size = { border_size, border_size },
    };

    pushArray(GpuUiRect, &state.gpu_ui_rects, rect);
}

void ed_init(_gfx_frame_func frame_func) {
    state.gfx_cx = gfx_init_context(frame_func, 640, 480);
    state.ui_cx = ui_init_context();

    float vertices[] = {
        // positions       texture coords
        -1.0f,  1.0f,       0.0f, 0.0f,
         1.0f,  1.0f,       1.0f, 0.0f,
         1.0f, -1.0f,       1.0f, 1.0f,
        -1.0f, -1.0f,       0.0f, 1.0f,
    };
    const uint16_t indices[] = { 0, 1, 2,  0, 2, 3 };

    // NOTE: the order of these matter
    gfx_push_vertex_buffer(state.gfx_cx, vertices, sizeof(vertices));
    gfx_push_vertex_buffer(state.gfx_cx, indices, sizeof(indices));

    state.gpu_ui_rects = newArray(GpuUiRect, 2000);
    state.gpu_glyphs = newArray(GpuGlyph, 1024);
    state.glyph_cache = newArray(GpuGlyph, 97);

    gfx_allocate_vertex_buffer(state.gfx_cx, state.gpu_glyphs.capacity * sizeof(GpuGlyph));
    gfx_allocate_vertex_buffer(state.gfx_cx, sizeof(GpuUniformParams));
    gfx_allocate_vertex_buffer(state.gfx_cx, state.gpu_ui_rects.capacity * sizeof(GpuUiRect));

    uint8_t ttf_buffer[1<<20];
    // TODO: grab default font from the system
    FILE *ttf_file = fopen("./bin/JetBrainsMono-Medium.ttf", "rb");
    if (!ttf_file) {
        fprintf(stderr, "failed to load font\n");
        exit(1);
    }
    assert(fread(ttf_buffer, 1, 1<<20, ttf_file));
    fclose(ttf_file);

    stbtt_fontinfo font;
    stbtt_InitFont(&font, ttf_buffer, stbtt_GetFontOffsetForIndex(ttf_buffer, 0));

    const int font_bitmap_size = 1024;
    const int rasterized_font_height = 64;
    uint8_t *font_bitmap = context_alloc(font_bitmap_size*font_bitmap_size * sizeof(uint8_t));

    int ascent, descent, line_gap;
    float scale = stbtt_ScaleForPixelHeight(&font, rasterized_font_height);
    stbtt_GetFontVMetrics(&font, &ascent, &descent, &line_gap);

    for (size_t xx=0; xx < rasterized_font_height/2; ++xx) {
        for (size_t yy=0; yy < rasterized_font_height; ++yy) {
            font_bitmap[xx + yy * font_bitmap_size] = 0;
        }
    }
    // manually add glyph for SPACE
    pushArray(GpuGlyph, &state.glyph_cache, ((GpuGlyph){
        .atlas_position = { 0 },
        .size = { rasterized_font_height/4, 1 },
        .position = { 0 },
        .y_offset = -rasterized_font_height,
    }));

    int x = rasterized_font_height/4;
    int y = 0;
    for (size_t i = 33; i < 33+96; ++i) {
        int width,height, xoff,yoff;
        uint8_t *bitmap = stbtt_GetCodepointBitmap(&font, scale, scale, (int)i, &width, &height, &xoff, &yoff);

        if (x + width >= font_bitmap_size) {
            x = 0;
            y += (int)((float)(ascent - descent + line_gap)*scale);
        }

        size_t xxx = x;
        for (size_t xx=0; xx < (size_t)width; ++xx) {
            size_t yyy = y;
            for (size_t yy=0; yy < (size_t)height; ++yy) {
                font_bitmap[xxx + yyy * font_bitmap_size] = bitmap[xx + yy * width];

                yyy += 1;
            }

            xxx += 1;
        }

        pushArray(GpuGlyph, &state.glyph_cache, ((GpuGlyph){
            .atlas_position = { (float)(x), (float)(y) },
            .size = { (float)width, (float)height },
            .position = { (float)x, (float)y },
            .y_offset = (float)(yoff),
        }));

        x += width;
    }

    gfx_push_texture_buffer(state.gfx_cx, font_bitmap_size, font_bitmap_size, font_bitmap, font_bitmap_size*font_bitmap_size * sizeof(uint8_t));

    state.test_buffer = ed_buffer_from_file(_String(""));
}

void render_ui_text(string text, float position[2]) {
    queue_text(text, position);
}
void render_ui_rect(float position[2], float size[2]) {
    queue_ui_rect(position, size, 0);
}

void ed_frame() {
    state.ui_cx.frame_elements.data[0].size.computed_size[0] = state.gfx_cx->frame_width;
    state.ui_cx.frame_elements.data[0].size.computed_size[1] = state.gfx_cx->frame_height;

    ui_element(&state.ui_cx, _String("Number 1"), UI_AXIS_VERTICAL, ui_make_size(ui_children_sum, ui_fill), 0);
    ui_push_parent(&state.ui_cx);
    {
        ui_element(&state.ui_cx, _String("ui element 2"), UI_AXIS_HORIZONTAL, ui_make_size(ui_fit_text, ui_fill), UI_FLAG_DRAW_BACKGROUND|UI_FLAG_DRAW_TEXT);
        ui_element(&state.ui_cx, _String("ui element 2"), UI_AXIS_HORIZONTAL, ui_make_size(ui_fit_text, ui_fit_text), UI_FLAG_DRAW_BACKGROUND|UI_FLAG_DRAW_TEXT);
    }
    ui_pop_parent(&state.ui_cx);

    // Test buffer
    ed_buffer_render(&state.ui_cx, &state.test_buffer);

    ui_compute_layout(&state.ui_cx, 0);

    state.gpu_glyphs.size = 0;
    state.gpu_ui_rects.size = 0;
    ui_render(&state.ui_cx, render_ui_text, render_ui_rect);

    ui_update_cache(&state.ui_cx, 0);
    ui_prune(&state.ui_cx);

    state.gfx_cx->num_glyphs = state.gpu_glyphs.size;
    if (state.gpu_glyphs.size > 0) {
        gfx_update_buffer(state.gfx_cx, 2, state.gpu_glyphs.data, state.gpu_glyphs.size * sizeof(GpuGlyph));
    }

    state.gfx_cx->num_ui_rects = state.gpu_ui_rects.size;
    if (state.gpu_ui_rects.size > 0) {
        gfx_update_buffer(state.gfx_cx, 4, state.gpu_ui_rects.data, state.gpu_ui_rects.size * sizeof(GpuUiRect));
    }

    GpuUniformParams gpu_uniform_params = {
        .screen_size = {
            (float)state.gfx_cx->frame_width,
            (float)state.gfx_cx->frame_height,
        },
    };

    gfx_update_buffer(state.gfx_cx, 3, &gpu_uniform_params, sizeof(GpuUniformParams));
}

int main(int argc, char* argv[]) {
    ed_init(ed_frame);

    while(keep_running) {
        gfx_run_events(state.gfx_cx);
    }
}
